def type_check(obj, valid_classes):
    if not isinstance(obj, valid_classes):
        raise ValueError(
            f"Expected `obj` to be type from {valid_classes}, but got type {type(obj)}"
            " instead"
        )



class SubwinderError(Exception):
    """
    This is the base exception that all the other custom exceptions are derived from.
    It's honestly just used when either something violates an assumption I made about
    the API.
    """


class SubLangError(SubwinderError):
    """
    Raised when the user tries to convert a language code using one of the global
    objects from `subwinder.lang` using an invalid value.
    """


class SubAuthError(SubwinderError):
    """
    Raised when trying to use `AuthSubwinder` without providing a username, password, or
    useragent. Or when the API returns a response indicating that user
    tried to perform an invalid action (401) or when the useragent isn't valid in some
    way (411, 414, 415).
    """


class SubUploadError(SubwinderError):
    """
    Raised if the user attempts to upload subtitles and the server returns a response
    for either invalid subtitles format (402) or internal subtitle validation failure
    (416).
    """


class SubDownloadError(SubwinderError):
    """
    Raised when the user tries to download subtitles in a way that needs either the
    directory name or media name without having been provided that information. Or this
    is raised when the user has reached the download limit.
    """


class SubServerError(SubwinderError):
    """
    When the server raises a 5xx error on a request or when the user hits more than 40
    requests in 10 seconds the library retries the request with exponential backoff for
    at least 10 seconds, but if that timeout is exceeded and the server is still
    returning either a 5xx or too many requests then a `SubServerError` is raised.
    """


class SubHashError(SubwinderError):
    """
    Raised when the user in some way tries to get the `subwinder.utils.special_hash` of
    a file that is below the minimum required size of 128 KiB.
    """


class SubLibError(SubwinderError):
    """
    Raised whenever there is an error based around the use of the library. This is
    primarily caused by invalid parameters getting passed through (no matter how strict
    I make most things).
    """



import base64
import gzip
from pathlib import Path

from subwinder.exceptions import SubHashError


def extract(bytes, encoding):
    """
    Extract `bytes` from being gzip'd and base64 encoded.
    """
    compressed = base64.b64decode(bytes)
    return gzip.decompress(compressed).decode(encoding)


# As per API spec with some tweaks to make it a bit nicer
# https://trac.opensubtitles.org/projects/opensubtitles/wiki/HashSourceCodes
def special_hash(filepath):
    """
    The "special hash" used by opensubtitles representing a specific media file.
    """
    CHUNK_SIZE_BYTES = 64 * 1024
    FILE_MIN_SIZE = CHUNK_SIZE_BYTES * 2

    # Force `filepath` to be `Path`
    filepath = Path(filepath)

    with filepath.open("rb") as file:
        filesize = filepath.stat().st_size
        hasher = _SumHasher(filesize)

        if filesize < FILE_MIN_SIZE:
            raise SubHashError(f"Filesize is below minimum of {FILE_MIN_SIZE} bytes")

        # Hash the first chunk of the file
        hasher.update(file.read(CHUNK_SIZE_BYTES))
        # and the last chunk
        file.seek(-CHUNK_SIZE_BYTES, 2)
        hasher.update(file.read(CHUNK_SIZE_BYTES))

    return hasher.digest()


class _SumHasher:
    def __init__(self, filesize):
        self.hash = filesize
        self.digest_size = 8
        self.block_size = 8

    def update(self, values):
        for i in range(0, len(values), self.block_size):
            chunk = values[i : i + self.block_size]
            self.hash += int.from_bytes(chunk, byteorder="little")

    def digest(self):
        # Keep output in `self.digest_size`
        temp = self.hash & 2 ** (self.digest_size * 8) - 1
        return f"{temp:016x}"
